import io
import os
import time

from pydatpiff.utils.filehandler import File, Tmp
from pydatpiff.utils.utils import Object, Select, ThreadQueue, threader_wrapper

from .backend.audio.player import Player
from .backend.mediasetup import Album, Mp3
from .constants import verbose_message
from .errors import MediaError
from .frontend import screen
from .mixtapes import Mixtape
from .urls import Urls
from .utils.request import Session

Verbose = screen.Verbose


class Media:
    """A media player that control the songs selected from Mixtape"""

    player = None

    def __init__(self, mixtape=None, pre_select=None, player="mpv", **kwargs):
        """
        Initialize a media player and load all mixtapes.

        Args:
            mixtape (instance class) -- pydatpiff.Mixtape instance (default: {None})
            pre_select (Integer,String) --  pre-selected mixtape's album, artist,or mixtapes.
                    See media.setMedia for more info (default: None - Optional)

        Raises:
            MediaError: Raises MediaError if mixtapes is not a subclass of pydatpiff.Mixtape.
        """
        self.__setup(player)
        # Check if mixtape is valid
        self.__is_valid_mixtape(mixtape)

        self._album_cover = None
        self.bio = None
        self._Mp3 = None
        self.url = None
        self.uploader = None
        self._session = Session()
        self.mixtape = mixtape
        self._artist_name = None
        self._album_name = None
        self._selected_song = None
        self.__cache_storage = {}
        self.AUTOPLAY_INACTIVITY_TIME = 60 * 5  # 5 minutes

        Verbose(verbose_message["MEDIA_INITIALIZED"])

        if pre_select:  # Run setMedia with argument here
            # This step is optional for users, but can save an extra setup
            # when selecting an album in setMedia.
            self.setMedia(pre_select)

    def __str__(self):
        album_name = self._album_name
        if album_name:
            return "{} Mixtape".format(album_name)
        return str(self.mixtape)

    def __len__(self):
        try:
            return len(self.songs)
        except MediaError:
            Verbose(verbose_message["MEDIA_NOT_SET"])
            return 0

    def __setup(self, player=None):
        if not hasattr(self, "__temp_file"):
            Tmp.remove_temp_file_on_startup()
            self.__temp_file = Tmp.create()

        # Set the initial player. fallback to mpv if not specified
        self.player = Player.getPlayer(player)

    def _select(self, choice):
        """
        Queue and load  a mixtape to media player.
                            (See pydatpiff.media.Media.setMedia)

        :param: choice - (int) user selection by indexing an artist name or album name
                            (str)
        """
        # Map user selection according to the incoming datatype.
        # We map an integer to an artist and str to a mixtape.
        selection = None

        if isinstance(choice, int):
            # Adjust the selection to the correct index.
            total_mixtape = len(self.mixtape)
            if choice >= total_mixtape:
                choice = total_mixtape
            selection = Select.get_leftmost_index(choice, options=self.mixtape.artists)
        else:
            options = self.mixtape.artists.copy()
            options.extend(self.mixtape.mixtapes)

            selection = Select.get_index_of(choice, options=options)
        assert selection is not None, "Invalid selection"
        return selection

    def setMedia(self, mixtape):  # noqa
        """
        Initialize and set the Mixtape to Media Player.
        A pydatpiff.mixtapes.Mixtape's album will be load to the media player.

        Args:
            selection - pydatpiff.Mixtape album's name or artist's name.
                int - will return the Datpiff.Mixtape artist at that index.
                str - will search for an artist from Mixtape.artists (default)
                    or album from Mixtape.album.

                See pydatpiff.mixtape.Mixtape for a mixtape album or artist selection
        """

        # Mixtape's class will handle errors and None value

        mixtape_index = min(self._select(mixtape), len(self.mixtape) - 1)

        # set Media's Mixtape object attributes
        url = self.mixtape._links[mixtape_index]  # noqa
        self.url = "".join((Urls.datpiff["base"], url))
        self.artist = self.mixtape.artists[mixtape_index]
        self.album_cover = self.mixtape.album_covers[mixtape_index]

        # set Media's Album detail attributes
        self.album = Album(url)
        self._Mp3 = Mp3(self.album)

        # get the album's uploader and bio
        self.uploader = getattr(self.album, "uploader", "unknown uploader")
        self.bio = getattr(self.album, "uploader", "no bio")
        formatted_title = " - ".join((self.artist, self.album.name))
        Verbose(verbose_message["MEDIA_SET"] % formatted_title)

    def __is_valid_mixtape(self, instance):
        """Verify media mixtape subclass is an instance of mixtapes' class

        Args:
            instance {instance class} -- Pydatpiff Mixtape instance

        Returns:
            Boolean -- True or False if instance is subclass of pydatpiff Mixtape class
        """
        if not instance:
            raise MediaError(1)

        if not issubclass(instance.__class__, Mixtape):
            raise MediaError(2, '"mixtape" must be Mixtape object')

    def find_song(self, name):
        """
         Search through all mixtapes songs and return all songs
         with song_name

        Args:
            song_name {str} -- song to search for.

        Returns:
            tuple -- returns a tuple containing mixtapes data (index,artist,album) from search.
        """

        # NOTE: Take a look at this video by James Powell
        # https://www.youtube.com/watch?v=R2ipPgrWypI&t=1748s at 55:00.

        song_name = Object.strip_and_lower(name)
        Verbose("\n" + verbose_message["SEARCH_SONG"] % song_name)
        links = self.mixtape.links
        links = list(enumerate(links, start=1))
        results = ThreadQueue(Album.lookup_song, links).execute(song=song_name)
        if not results:
            Verbose(verbose_message["SONG_NAME_NOT_FOUND"] % song_name)
        results = Object.remove_list_null_value(results)
        return results

    def _index_of_song(self, select):
        """
        Parse all user input and return the correct song index.
        :param select: -  Media.songs name or index of Media.songs
               datatype: int: must be numerical
                         str: artist,mixtape, or song name
        """
        try:
            if isinstance(select, int):
                return Select.get_leftmost_index(select, self.songs)
            return Select.get_index_of(select, self.songs)
        except ValueError:
            raise MediaError(5)

    @property
    def artist(self):
        """Return the current artist name."""
        if not hasattr(self, "_artist_name"):
            self._artist_name = None
        return self._artist_name

    @artist.setter
    def artist(self, name):
        self._artist_name = name

    @property
    def album(self):
        """Return the current album name."""
        if not hasattr(self, "_album_name"):
            self._album_name = None
        return self._album_name

    @album.setter
    def album(self, name):
        self._album_name = name

    @property
    def album_cover(self):
        if hasattr(self, "_album_cover"):
            return self._album_cover

    @album_cover.setter
    def album_cover(self, url):
        self._album_cover = url

    @property
    def songs(self):
        """Return all songs from album."""
        if self._Mp3 is None:
            extra_message = '\nSet media by calling -->  Media.setMedia("album-name")'
            raise MediaError(3, extra_message)
        return self._Mp3.songs

    @property
    def mp3_urls(self):
        """Returns all parsed mp3 url"""
        return list(self._Mp3.mp3_urls)

    def show_songs(self):
        """Pretty way to Print all song names"""
        try:
            for index, song in enumerate(self.songs, start=1):
                Verbose("%s: %s" % (index, song))
        except MediaError:
            Verbose(verbose_message["MEDIA_NOT_SET"])

    @property
    def song(self):
        """Returns the current song set by user."""
        return self._selected_song

    @song.setter
    def song(self, name):
        """
        Set the current song to the song name or index.
        :param name: - song name or index of song
        """
        try:
            index = self._index_of_song(name)
            self._selected_song = self.songs[index]
        except (ValueError, MediaError):
            Verbose(verbose_message["SONG_NAME_NOT_FOUND"] % name)

    def _cache_song(self, song, content):
        """
        Preserve the data from song and store it for future calls.
         This prevents calling the requests function again for the same song.
         Each data from a song will be stored in __cache_storage for future access.

        Args:
            song (str):  name of song
            content (byte): song's audio content
        """
        song = "-".join((self.artist, song))
        try:
            self.__cache_storage[song] = content
        except MemoryError:
            self.__cache_storage = {}

    def _retrieve_song_from_cache(self, song):
        """Retrieve song's audio content from cache
        Args:
            song (str): name of song

        Returns:
            Http response : A http response containing the song's audio contents.
        """
        requested_song = "-".join((self.artist, song))
        if requested_song in self.__cache_storage:
            return self.__cache_storage.get(requested_song)

    def _write_audio(self, track):
        """Write mp3 audio content to IO Bytes stream.
        Args:
            track (int,string): Name or index of song.

        Returns:
            BytesIO: A file-like API for reading and writing bytes objects.
        """

        try:
            selection = self._index_of_song(track)
            if selection is None:
                raise MediaError("Song not found")

            link = self.mp3_urls[selection]
            song_name = self.songs[selection]
            self._song_index = selection
        except:  # noqa
            return

        self.song = selection + 1

        # Retrieve song's content in cached
        response = self._retrieve_song_from_cache(song_name)
        if not response:
            response = self._session.method("GET", link)
            self._cache_song(song_name, response)

        return io.BytesIO(response.content)

    @property
    def autoplay(self):
        """Continuously play song from current mixtape album."""
        if hasattr(self, "_auto_play"):
            self.player._media_autoplay = self._auto_play
            return self._auto_play

    @autoplay.setter
    def autoplay(self, auto=False):
        """Sets the autoplay function.

        auto - disable or enable autoplay
                 datatype: boolean
                 default: False
        """
        self._auto_play = auto  # noqa
        self._continuous_play()
        if auto:
            Verbose(verbose_message["AUTO_PLAY_ENABLED"])
        else:
            Verbose(verbose_message["AUTO_PLAY_DISABLED"])

    @property
    def _is_autoplay_inactive(self):
        is_paused = self.player.state.get("paused")
        if is_paused and self._inactive_time == 0:
            self._inactive_time = time.time()
        return is_paused and time.time() - self._inactive_time > self.AUTOPLAY_INACTIVITY_TIME

    @threader_wrapper
    def _continuous_play(self):
        """
        Automatically play each song from Album when autoplay is enable.
        """

        if self.song is None:
            current_track = 1
            self.play(1)
        else:
            current_track = self._index_of_song(self.song)

        if not self.player.state["playing"]:
            current_track + 1

        while self.autoplay:
            """
            If the user di not select a track
            then set and play the first song from mixtape
            """
            # Handle autoplay pause inactivity
            if self._is_autoplay_inactive:
                self._inactive_time = 0
                Verbose(verbose_message["AUTO_PLAY_INACTIVITY"])
                self.autoplay = False
                self.player.stop
                break

            # handle track stoppages by system
            if self.player._state.get("system_stopped"):  # noqa:
                self.player.state["system_stopped"] = False
                next_track = current_track + 1
                if next_track > len(self):
                    Verbose(verbose_message["AUTO_PLAY_LAST_SONG"])
                    self.autoplay = False
                    break

                Verbose(verbose_message["AUTO_PLAY_NEXT_SONG"])
                self.play(next_track)

            # handle stoppage by user
            elif self.player.state.get("stopped"):
                break

    def _get_audio_track(self, track):
        """
        Perform a lookup for song by its index or name and returns
        full track name and audio content.

        Args:   track (int,string): Name or index of song.
        Returns:    tuple: (track name, audio content)
        """
        if track is None:
            Verbose("\n\t", verbose_message["NO_SONG_SELECTED"])
            raise MediaError(8, verbose_message["NO_SONG_SELECTED"])

        try:
            if isinstance(track, int):
                if track > len(self):
                    track = len(self)
            elif isinstance(track, str):
                track = self._index_of_song(track) + 1
            else:
                raise ValueError
        except (MediaError, ValueError):  # ^ will throw ValueError if track  name is invalid
            Verbose(verbose_message["SONG_NAME_NOT_FOUND"] % track)
            raise MediaError(8, verbose_message["SONG_NAME_NOT_FOUND"] % track)

        content = self._write_audio(track)
        if not content:
            Verbose(verbose_message["UNAVAILABLE_SONG"])
            raise MediaError(9, verbose_message["UNAVAILABLE_SONG"])

        song_name = self.songs[self._song_index]
        return song_name, content.read()

    def play(self, track=None, demo=False):
        """Play selected mixtape's track

        Args:
            track (int,string)- name or index of song.
            demo (bool, options) - True: demo buffer of song (default: False).
                False: play full song
        """
        try:
            song_name, content = self._get_audio_track(track)
        except MediaError:
            return

        buffer_size = len(content)
        # play demo or full song
        if not demo:  # demo whole song
            chunk = content
            buffer = int(buffer_size)
        else:  # demo partial song
            buffer = int(buffer_size / 5)
            start = int(buffer / 5)
            chunk = content[start : buffer + start]
        size = File.get_human_readable_file_size(buffer)

        # write song to file
        File.write_to_file(self.__temp_file.name, chunk, mode="wb")

        # display message to user
        screen.display_play_message(self.artist, self.album, song_name, size, demo)

        song = " - ".join((self.artist, song_name))
        self.player.set_track(song, self.__temp_file.name)
        self.player.play  # noqa - play song is a property of the player class

    def download(self, track=None, rename=None, output=None):
        """
        Download song from Datpiff

        Args:
            track (int,string) - name or index of song type(str or int)
            output (string) - location to save the song (optional)
            rename (string) - rename the song (optional)
                default will be song's name
        """
        try:
            song, content = self._get_audio_track(track)
        except MediaError:
            # Exception message will be handled from by `_get_audio_track`
            return

        # Handles paths
        output = output or os.getcwd()
        if not File.is_dir(output):
            raise FileNotFoundError("Invalid directory: %s" % output)

        # Handles song's renaming
        if rename:
            song, _ = os.path.splitext(rename)
        title = " - ".join((self.artist, song.strip() + ".mp3"))

        file_name = File.standardize_file_name(title)
        file_name = File.join(output, file_name)

        size = File.get_human_readable_file_size(len(content))
        File.write_to_file(file_name, content, mode="wb")
        screen.display_download_message(title, size)

    def download_album(self, output=None):
        """Download all tracks from Mixtape.

        Args:
            output ([type], optional): path to save mixtape.(default: current directory)
        """
        if not output:
            output = os.getcwd()
        elif not os.path.isdir(output):
            Verbose(verbose_message["INVALID_DIRECTORY"] % output)
            return

        album_dirname = " - ".join((self.artist, self.album.name))
        output = os.path.join(output, File.standardize_file_name(album_dirname))
        os.mkdir(output)
        # must follow `download` method's arguments order
        ThreadQueue(
            self.download,
            self.songs,
        ).execute(rename=None, output=output)
        Verbose("\n" + verbose_message["SAVE_ALBUM"] % (self.artist + " " + self.album.name, output))



import logging
import warnings

import requests
from requests.adapters import HTTPAdapter

from pydatpiff.errors import RequestError

# Set Request logging levels
logging.getLogger("requests").setLevel(logging.CRITICAL)
logging.getLogger("urllib3").setLevel(logging.CRITICAL)


class Session:  # pragma: no cover
    """Dynamic way to keep requests.Session throughout whole programs."""

    # private
    _TOTAL_TIMEOUT = 0
    _MAX_RETRIES = 2
    _CACHE = {}

    # public
    TIMEOUT = 3
    session = requests.Session()

    def __init__(self, *arg, **kwargs):
        transport_adapter = HTTPAdapter(pool_connections=50, pool_maxsize=50, max_retries=self._MAX_RETRIES)
        self.session.mount("https://", transport_adapter)

    @classmethod
    def put_in_cache(cls, url, response):
        try:
            url = url.strip()
            cls._CACHE[url] = response
        except MemoryError:
            cls.clear_cache()

    @classmethod
    def clear_cache(cls):
        """clear _CACHE to prevent memory error"""
        del cls._CACHE
        cls._CACHE = {}

    def get_from_cache(self, url):
        """Checks if url already have a response.
        Stop from calling the request method more than once.
        Great for saving mobile data on mobile devices.
        """
        url = str(url).strip()
        if url in self._CACHE.keys():
            return self._CACHE[url]

    def method(self, method, url, bypass=None, **kwargs):
        """urllib requests method"""
        valid_method = ["get", "post"]
        method = str(method).lower()

        if method not in valid_method:
            return

        cached_response = self.get_from_cache(url)
        if cached_response and method != "post":
            return cached_response

        try:
            # GET
            if method == "get":
                web = self.session.get(url, timeout=self.TIMEOUT, **kwargs)
            # POST
            if method == "post":
                web = self.session.post(url, timeout=self.TIMEOUT, **kwargs)

        except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as e:
            # first catch server connect error, then user's internet error
            if e is requests.exceptions.ReadTimeout:
                raise RequestError(1)

            # catch user's connection error
            self._TOTAL_TIMEOUT += 1
            if self._TOTAL_TIMEOUT >= 3:
                print("\n")  # need for spacing
                warn_msg = "\nWarning: Please check your internet connection !"
                warnings.warn(warn_msg)
                self._TOTAL_TIMEOUT = 0
            raise RequestError(2)

        except requests.exceptions.InvalidURL:
            raise RequestError(3)

        # process the request for HTTP Errors
        try:
            web.raise_for_status()
        except:
            raise RequestError(4)
        else:
            # cache the request response for later use cases
            self.put_in_cache(url, web)
            self._TOTAL_TIMEOUT = 0
        finally:
            return web



import logging
import math
import os
import re
import tempfile

logger = logging.getLogger(__name__)


class Tmp(object):  # pragma: no cover
    """
    Handle all temporary files created by the media player.

    It removes the hanging temporary files that are not destroyed.
    Since we are creating temporary files to store mp3 content, we need
    to remove these files manually after being used.
    python3 module 'signal' sometimes fail to catch  signal
    on certain platforms like Windows IDLE.
    We create our own, in case user close window or program dies before
    being able to remove these files.
    """

    @staticmethod
    def create():
        """Create a temporary file with the suffix  `__datpiff`"""
        return tempfile.NamedTemporaryFile(suffix="_datpiff", delete=False)

    @staticmethod
    def remove_temp_file_on_startup():
        """remove all temporary file created by Datpiff on start up"""
        # every tmp file using media player will have this suffix
        suffix = "_datpiff"
        tmp_dir = tempfile.gettempdir()
        for filename in os.listdir(tmp_dir):
            if suffix in filename:
                try:
                    filename = os.path.join(tmp_dir, filename)
                    os.remove(filename)
                except FileNotFoundError:
                    logger.warning("File not found: %s", filename)


class File:
    @staticmethod
    def is_dir(path):
        path = path or ""
        return os.path.isdir(path)

    @staticmethod
    def is_file(path):
        return os.path.isfile(path)

    @classmethod
    def join(cls, path=None, to=""):
        path = path or os.getcwd()
        if not cls.is_dir(path):
            raise FileNotFoundError(f"{path} is not a directory")
        return os.path.join(path, to)

    @staticmethod
    def standardize_file_name(name):
        return re.sub(r"[^\w_\s\-.]", "", name)

    @classmethod
    def write_to_file(cls, filename, content, mode="wb"):
        with open(filename, mode) as f:
            f.write(content)

    @staticmethod
    def get_human_readable_file_size(buf_size):
        """Convert file size and returns user readable size"""
        if buf_size == 0:
            return "0B"

        ext = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
        change = int(math.floor(math.log(buf_size, 1024)))
        power = math.pow(1024, change)
        calc = buf_size / power
        file_size = round(calc, 2) if buf_size > 999 else int(calc)
        return "%s%s" % (file_size, ext[change])



import logging  # noqa
import logging.config
import os

import yaml

DIR = os.path.dirname(os.path.abspath(__file__))
log_config_file = os.path.join(DIR, ".logging_config.yaml")

with open(log_config_file, "r") as f:
    config = yaml.load(f.read(), Loader=yaml.FullLoader)
    logging.config.dictConfig(config)



import re

from pydatpiff.constants import SERVER_DOWN_MSG
from pydatpiff.errors import DatpiffError, Mp3Error
from pydatpiff.urls import Urls
from pydatpiff.utils.request import Session
from pydatpiff.utils.utils import Object

from .scraper import MediaScraper


class DatpiffPlayer:
    """
    Datpiff's frontend media player object
    Note: This class is not a stand-alone class
    """

    # Flag for mobile's version.
    # Fallback if desktop version is not working ('website issue')
    _USE_MOBILE_VERSION = False

    def __init__(self, *args, **kwargs):
        base_method = "lookup_song"  # required methods
        if not hasattr(self, base_method):
            raise NotImplementedError("DatpiffPlayer not a stand-alone class")
        super().__init__(*args, **kwargs)

    def _verify_version(self):
        try:
            self.name = (
                re.search(
                    r"class=\"title\">(.*)<",
                    self.embedded_player_content,
                )
                .group(1)
                .strip()
            )
        except AttributeError:
            raise Mp3Error(3, "Could not find album's name")
        return self.name

    @property
    def embedded_player_content(self):
        """Returns Datpiff embedded player response text"""
        # Note: Request Sessions are being cached for every request.
        #      If the url endpoint is found in the cached, the request
        #      will NOT be recalled.  Instead, the cached response will be returned.
        url = self.build_web_player_url(self._album_ID)
        try:
            return self._session.method("GET", url).text
        except:  # noqa
            raise DatpiffError(1, SERVER_DOWN_MSG)

    def _check_datpiff_version(self):
        """
        function that will check program and determine
         which version ( desktop or mobile ) to use.

         As of July 10, 2020, Datpiff's desktop's version is broken,
         and is not populating album data.

         Data that is NOT being populated are ONLY the followings:
           - Album.name
           - Mp3.songs

         All other functions are still working as expected.

         This function will check if an album name is populated correct.
         if not then mobile version will be used as a fallback
        """
        # we check if Album.name attribute exists.
        # If it doesn't, we switch to mobile version

        if self._verify_version() is None:
            self._USE_MOBILE_VERSION = True

    @classmethod
    def build_web_player_url(cls, album_id):
        """Creates url link for Datpiff's embedded music player."""

        # July 10, 2020 , This will fix error with songs name not populating
        # if desktop version fails, flag program to use Mobile version as a fallback
        version = "mobile" if cls._USE_MOBILE_VERSION else "embeds"
        return "".join(
            (
                "https://{}.datpiff.com/mixtape/".format(version),
                str(album_id),
                "?trackid=1&platform=desktop",
            )
        )


class Album(DatpiffPlayer):
    """
    Renders Datpiff's Mixtape page and create URI link to its media player object.
    Data from URI link will be process and use to populate data for mixtapes. This data
    includes:
        Album uploader name and bio
        Album's name and songs
    """

    _session = Session()

    def __init__(self, link):
        """
        Media player Album object constructor.
        :param link: Link to the media player page.
        """

        self._name = None
        self.link = "".join((Urls.datpiff["album"], link))
        self._check_datpiff_version()

    def __str__(self):
        return self.name

    @property
    def name(self):
        # for desktop version issue we will use the mobile version
        if hasattr(self, "_name"):
            return self._name

    @name.setter
    def name(self, album):
        self._name = album

    @property
    def _album_ID(self):
        """Album ID Number"""
        return MediaScraper.get_album_suffix_number(self.link)

    @property
    def bio(self):
        return MediaScraper.get_uploader_bio(self.embedded_player_content)

    @property
    def _album_html(self):
        """
        Return the requests' response from the current Mixtape link
            See __init__ or mixtapes.Mixtape.links.
        """
        # Session responses are cached,
        # so we don't have to worry about recalling requests.
        response = self._session.method("GET", self.link)
        if response:
            return response.text
        return " "

    @property
    def uploader(self):
        return MediaScraper.get_uploader_name(self._album_html)

    @classmethod
    def lookup_song(cls, links, song, *args, **kwargs):
        """
        Search through all Albums and return all Albums
        that contains similar songs' title.

        Args:
                 song (string) - title of the song to search for
                 links (tuple) -  index of mixtape link and mixtape link
        """
        index, link = links
        album = cls(link)
        tracks = Mp3(album).songs
        for track in tracks:
            song = Object.strip_and_lower(song)
            if song in Object.strip_and_lower(track):
                return {"index": index, "album": album.name, "song": track}


class Mp3:
    def __init__(self, album):
        if not getattr(album, "embedded_player_content", None):
            raise Mp3Error(1, "No album response found")

        self.album = album
        self.album_response = album.embedded_player_content

    def __len__(self):
        if self.songs:
            return len(self.songs)
        return 0

    def __str__(self):
        if getattr(self, "album"):
            return " ".join((str(self.album), "Mp3"))
        return "MP3"

    @property
    def songs(self):
        """Returns all songs name from album."""
        return MediaScraper.get_song_titles(self.album_response)

    @property
    def __urlencoded_tracks(self):
        """Url encode audio url"""
        songs = MediaScraper.get_mp3_urls(self.album_response)
        return [re.sub(r"\s", "%20", song) for song in songs]

    @property
    def _album_id(self):
        """Media Album reference ID number Ex: 6/m1393dba"""
        return MediaScraper.get_embed_player_id(self.album_response)

    @property
    def mp3_urls(self):
        url = "https://hw-mp3.datpiff.com/mixtapes/"
        for track in self.__urlencoded_tracks:
            endpoint = "{}{}".format(self._album_id, track)
            yield "".join((url, endpoint))



import re

from mutagen.mp3 import MP3

from pydatpiff.constants import music_symbols, player_state_keys
from pydatpiff.errors import PlayerError
from pydatpiff.frontend.screen import Verbose
from pydatpiff.utils.utils import threader_wrapper


class BaseMeta(type):
    """
    For Authors and Contributors only !
    Metaclass that will force constrains on any subclass that inherit BasePlayer

    The below functions and methodology must be implemented along with subclass.
    They will be forced here in BaseMeta(MetaClass). See: BaseMeta.methods
    """

    # more variables: _songs,current_time

    def __new__(cls, name, bases, attrs):
        methods = [
            "set_track",
            "duration",
            "current_time",
            "_format_time",
            "_seeker",
            "volume",
            "volume_up",
            "volume_down",
            "play",
            "pause",
            "rewind",
            "ffwd",
            "stop",
        ]
        # NOTE: 10/28/2021 - need:  See vlc and mpv player
        #    "_global_volume" - to control initial volume on media switch
        #     "volume_level" - setter and getter
        # more variables: _songs,current_time
        cls._name = name
        cls._bases = bases
        cls._attrs = attrs

        for _ in bases:
            for method in methods:
                if method not in attrs:
                    error = 'Method: "%s.%s" must be implemented in your Derived Class ' "to use BasePlayer" % (
                        name,
                        method,
                    )

                    raise NotImplementedError(error)
        return super().__new__(cls, name, bases, attrs)


class BasePlayer(metaclass=BaseMeta):
    """Media player base controller"""

    _global_volume = 100
    _is_monitoring = False
    _track_start_time = 0

    """
    Set default player state
        E.g.
            "playing": False,
            "paused": False,
            "stopped": False,
            "system_stopped": False,
        }
    """
    # NOTE: do not change `_state` private variables to public variables
    _state = dict((k, False) for k in player_state_keys)

    def __init__(self, *args, **kwargs):
        self._track_loaded = False
        self._track_playing = False
        self._track_paused = False
        self._track_stopped = False
        self._system_stopped = False

    def __len__(self):
        # duration will be forced to be implemented by Meta class
        return int(self.duration)

    def reset_and_update_state(self, update=None):
        """Reset and update  all player's state"""
        self._state.update(playing=False, paused=False, loaded=False, stopped=False)
        #  update state if specified.
        update = update if update else {}
        self._state.update(**update)

    @property
    def name(self):
        """Get the current track title"""
        if not hasattr(self, "_song"):
            raise PlayerError(2, "Cannot find song's name ")
        return self._song

    @property
    def state(self):
        """Current state of the song being played"""
        if hasattr(self, "_state"):
            return self._state
        else:
            raise PlayerError(3, "def _state")

    @state.setter
    def state(self, state):
        self._state = state

    @property
    def _track_loaded(self):
        """Return player loaded state"""
        return self.state["loaded"]

    @_track_loaded.setter
    def _track_loaded(self, state=False):
        self.state["loaded"] = bool(state)

    @property
    def _track_playing(self):
        """Return the track playing state"""
        # if boolean param not specified, then return the playing state
        return self.state["playing"]

    @_track_playing.setter
    def _track_playing(self, state=False):
        """
        Set the state of playing and pause.

        param: boolean - True or False
                True: sets playing True and pause to False.
                False: sets playing False and pause True.
        """
        self.state["playing"] = state

    @property
    def _track_paused(self):
        return self.state["paused"]

    @_track_paused.setter
    def _track_paused(self, state=False):
        self.state["paused"] = bool(state)

    @property
    def _track_stopped(self):
        return self.state["paused"]

    @_track_stopped.setter
    def _track_stopped(self, state=False):
        self.state["stopped"] = bool(state)

    @property
    def _system_stopped(self):
        return self.state["paused"]

    @_system_stopped.setter
    def _system_stopped(self, state=False):
        self.state["system_stopped"] = bool(state)

    @threader_wrapper
    def auto_manage_state(self, *args, **kwargs):
        DEBOUNCE_TIME = 1000 * 2  # two seconds
        state = dict(
            loaded=False,
            playing=False,
            paused=False,
            stopped=False,
            system_stopped=False,
        )
        while True:

            # Handle system stop
            # Automatically stop track - mainly used for media.autoplay feature.
            if self.current_time > 0 and self.current_time >= (self.duration - DEBOUNCE_TIME):
                self.stop
                state.update(dict(stopped=False, system_stopped=True))

            if self._track_loaded and not self._track_playing:
                if self.current_time > 0:
                    state.update(dict(loaded=True, paused=True))
                else:
                    state.update(dict(loaded=True))

            elif self._track_playing:
                state.update(dict(loaded=True, playing=True))

            elif self._track_paused:
                state.update(dict(loaded=True, paused=True))

            elif self._track_stopped:
                self.stop
                self._paused_time = 0
                self._track_start_time = 0
                state.update(dict(system_stopped=True))

            self.state = state
            if self.state.get("stopped", True):
                break

    def set_track(self, *args, **kwargs):
        # Media class method needed on all player
        raise NotImplementedError

    @property
    def duration(self, *args, **kwargs):
        raise NotImplementedError

    def _format_time(self, *args, **kwargs):
        """Format current song time to clock format"""
        raise NotImplementedError

    @property
    def info(self):
        """Current state of the song"""
        if not hasattr(self, "_song") or not hasattr(self, "current_time"):
            return "no song loaded"

        if "vlc" in str(self.__class__.__name__).lower():
            # VLC player has its own state manager
            state = re.match(r"[\w.]*\.(\w*)", str(self._player.get_state())).group(1)
        else:
            state = self._state

        symbol = "[]"
        if isinstance(state, str) and state != "NothingSpecial":
            # convert state's str to key/value pairs.
            # On initialization, set all state values => False
            vlc_state = state.lower()
            # Reset state
            for k, v in self._state.items():
                self._state[k] = False

            # See vlc.get_state docs for more info on VLC states
            self._state["playing"] = "playing" in vlc_state
            self._state["paused"] = "paused" in vlc_state
            self._state["stopped"] = "ended" in state.lower() or "error" in vlc_state or "stopped" in vlc_state
            self._state["system_stopped"] = "error" in vlc_state

        """
        NOTE: we don't want to reset all state, we only want to update the state and its opposing state
        e.g: playing: True -> paused: False
             paused:True -> playing: False

             stopped: True -> playing: False
             playing: True -> stopped: False
        """

        if self._state.get("paused"):
            self._track_paused = True
            self._track_playing = False
            symbol = music_symbols["paused"]
        elif self._state.get("playing"):
            self._track_playing = True
            self._track_paused = False
            symbol = music_symbols["playing"]
        else:
            self.reset_and_update_state()
            symbol = music_symbols["stopped"]

        # song current time
        current_min, current_sec = self._format_time(self.current_time)
        # song duration
        duration_min, duration_sec = self._format_time(self.duration)

        current_sec = str(current_sec).zfill(2)
        duration_sec = str(duration_sec).zfill(2)
        track_info = "{0}  {1}:{2} - {3}:{4}\n".format(symbol, current_min, current_sec, duration_min, duration_sec)

        is_auto_playing = getattr(self, "_media_autoplay", False)
        autoplay_symbol = music_symbols["autoplay"] if is_auto_playing else ""
        Verbose("\n%s TRACK: %s" % (music_symbols["music"], self.name))
        Verbose(autoplay_symbol, track_info)

    @property
    def volume_level(self):
        """Current media player volume"""
        # TODO implement method to monitor the volume
        return 100

    @volume_level.setter
    def volume_level(self, level):
        """set Current media player volume"""
        self._global_volume = level
        # individual player volume control method here

    def volume_up(self, vol=5):
        """Turn the media volume up"""
        raise NotImplementedError

    def volume_down(self, vol=5):
        """Turn the media volume down"""
        raise NotImplementedError

    def volume(self, vol=None):
        """Set the volume to exact number"""
        raise NotImplementedError

    @property
    def play(self):
        """Play media song"""
        raise NotImplementedError

    @play.setter
    def play(self, *args, **kwargs):
        raise NotImplementedError

    @property
    def pause(self):
        """Pause the media song"""
        raise NotImplementedError

    def _seeker(self, pos=10, rew=True):
        """Rewind and ffwd base controls"""
        raise NotImplementedError

    def rewind(self, pos=10):
        """
        Rewind the media song
             @params: pos:: time(second) to rewind media. default:10(sec)
        """
        raise NotImplementedError

    def ffwd(self, pos=10):
        """
        Fast-forward the media song
            @params: pos:: time(second) to rewind media. default:10(sec)
        """
        raise NotImplementedError

    @property
    def stop(self):
        """Stops the song"""
        raise NotImplementedError


class MetaData(MP3):
    def __init__(self, track):
        super().__init__(track)

    @property
    def track_duration(self):
        return self.info.length



import atexit
import logging
import os
import platform
import subprocess

from pydatpiff.errors import PlayerError
from pydatpiff.utils.utils import threader_wrapper

logger = logging.getLogger(__name__)


class Popen(subprocess.Popen):  # pragma: no cover
    registered_popen = []
    _player_PID = None

    def __init__(self, *args, **kwargs):
        """build subprocess Popen object"""
        kwargs["stdin"] = subprocess.PIPE
        kwargs["stdout"] = subprocess.PIPE
        kwargs["stderr"] = subprocess.PIPE
        atexit.register(self.kill_on_quit)
        try:
            super().__init__(shell=False, *args, **kwargs)
            self._player_PID = self.pid
        except:
            logger.exception("PlayerPopenFailed")
            # throws PlayerNotFoundError
            raise PlayerError(6)

    def stop_player(self):
        if platform.system() == "Windows":
            # https://tweaks.com/windows/39559/kill-processes-from-command-prompt/
            kill_cmd = ("taskkill /pid {} /F",)
        else:
            # Systems: Linux, Darwin ..etc
            kill_cmd = "kill -9 {}"

        try:  # For Linux device, Mac, Ubuntu, Debain...etc
            if self._player_PID:
                return subprocess.check_call(
                    kill_cmd.format(self._player_PID),
                    shell=True,
                    stderr=subprocess.PIPE,
                )
        except subprocess.CalledProcessError:
            return os.kill(int(self._player_PID), 9)  # kill mpv using os
        except:
            logger.exception("Failed to kill player")

    @threader_wrapper
    def register(self, callback=None, *args, **kwargs):
        """
        Kills subprocess Popen when error occur or when process job finish

        Args:
            callback (function, optional): Function to execute once player process dies.
        """
        self.registered_popen.append(self)

        # constantly probe Popen to check if player is running
        while True:
            if self.poll() is not None:
                if callback:
                    callback(*args, **kwargs)
                self.kill()
                break

    @classmethod
    def unregister(cls):
        """Unregister and terminate Popen process"""
        for popen in cls.registered_popen:
            popen.kill()

    @property
    def is_alive(self):
        if self.poll() is None:
            return True
        return False

    def kill_on_quit(self):
        if self.is_alive:
            self.stop_player()

    def kill_on_start(self):
        for process in self.registered_popen:
            process.kill()



import vlc

from pydatpiff.errors import PlayerError
from pydatpiff.frontend.screen import Verbose

from .baseplayer import BasePlayer


class VLCPlayer(BasePlayer):
    def __init__(self, *args, **kwargs):
        try:
            self._vlc = vlc.Instance("-q")
            self._player = self._vlc.media_player_new()
        except:
            extended_msg = "Please check if your device supports VLC"
            raise PlayerError(1, extended_msg)

        self._volume = self._global_volume
        super().__init__(*args, **kwargs)

    @property
    def duration(self):
        return self._player.get_length()

    def _format_time(self, pos=None):
        """Format current song time to clock format"""
        mins = int(pos / 60000)
        secs = int((pos % 60000) / 1000)
        return mins, secs

    @property
    def current_time(self):
        return self._player.get_time()

    def set_track(self, name, path=None):
        if path:
            self._song = name
            self._path = path
            self._player.set_mrl(path)
            self._volume = self._global_volume
            self.auto_manage_state()
        else:
            Verbose("No media to play")

    @property
    def _volume(self):
        """Get current media player volume"""
        return self._player.audio_get_volume()

    @_volume.setter
    def _volume(self, level):
        """Set media player volume"""
        if level < 0:
            level = 0
        if level > 100:
            level = 100

        self._player.audio_set_volume(level)
        self._global_volume = level

    def volume_up(self, level=5):
        """Turn the media volume up"""
        if not level and not isinstance(level, int):
            return
        self._volume += level

    def volume_down(self, level=5):
        """Turn the media volume down"""
        if not level and not isinstance(level, int):
            return
        self._volume -= level

    def volume(self, level):
        """Set the volume to exact number"""
        if not level or not isinstance(level, int):
            return
        self._volume = level

    @property
    def play(self):
        """Play media song"""
        if not self._track_stopped:
            if self._track_paused:
                # unpause if track is already playing but paused
                self.pause
            else:
                self._player.play()

            self._track_playing = True
        else:
            try:
                self.stop
                self.set_track(self._song, self._path)
                self.state["stopped"] = False
                return self.play
            except RecursionError:
                self.state["stopped"] = True

    @property
    def pause(self):
        """Pause the media song"""

        is_paused = self._track_paused
        self._track_playing = is_paused
        self._player.pause()
        self._track_paused = not is_paused

    def _seeker(self, pos=10, rew=True):
        if self._state["stopped"]:
            return
        if rew:
            to_position = self._player.get_time() - (pos * 1000)
            # seeking far before track starts
            if to_position < 0:
                to_position = 0
        else:
            to_position = self._player.get_time() + (pos * 1000)
            # seeking too far beyond track ends
            if to_position > self.duration:
                to_position = self.duration - 1

        self._player.set_time(to_position)

    def rewind(self, pos=10):
        """Rewind track
        @params: pos:: time(second) to rewind media. default:10(sec)
        """
        self._seeker(pos, True)

    def ffwd(self, pos=10):
        """Fast forward track
        vlc time is in milliseconds
        @params: pos:: time(second) to rewind media. default:10(sec)
        """
        self._seeker(pos, False)

    @property
    def stop(self):
        self._player.stop()
        self._track_stopped = True



from pydatpiff.errors import InstallationError, PlayerError
from pydatpiff.utils.utils import Object

from .mpvplayer import MPV
from .vlcplayer import VLCPlayer


class Player:
    @classmethod
    def getPlayer(cls, player=None):

        player_options = {"vlc": VLCPlayer, "mpv": MPV}
        # return the player specified by the user
        try:
            if player:
                player = Object.strip_and_lower(player)
                selected_player_class = player_options.get(player, None)
                return selected_player_class.__call__()
        except:  # noqa
            extended_msg = "\nThe player you have chosen is not compatible with your device.\n"
            raise PlayerError(5, extended_msg)

        # if no player is specified by the user, then select a default player
        return cls._getDefaultPlayer()

    @classmethod
    def _getDefaultPlayer(cls):
        # Note: Player order will be respected!
        default_players = [MPV, VLCPlayer]
        for player in default_players:
            try:
                return player.__call__()
            except:  # noqa
                pass

        raise InstallationError(1)



"""
   Custom printing function using logger.
"""
from pydatpiff.utils.logging import logging

logger = logging.getLogger(__name__)


def Verbose(*args):  # noqa
    output = " ".join(args)
    logger.info(output)


def display_play_message(artist, album_name, song_name, size, demo=False):
    if demo:
        Verbose("\n%s %s %s" % ("-" * 20, "DEMO", "-" * 20))
    Verbose("\nAlbum: %s" % album_name)
    Verbose("Artist: %s" % artist)
    Verbose("Song: %s" % song_name)
    Verbose("Size:", size)


def display_download_message(song_name, size):
    Verbose("\nDownloading:", song_name, "\nSIZE:  ", size)



import re
from functools import wraps
from time import sleep, time

from pydatpiff.errors import MvpError
from pydatpiff.frontend.screen import Verbose
from pydatpiff.utils.filehandler import File

from ...utils.utils import threader_wrapper
from .audio_engine import Popen
from .baseplayer import BasePlayer, MetaData


class MPV(BasePlayer):
    # 10 mins:  Time to kill inattentive threads see pause.
    # Save the user's CPU :-)
    _MAX_INACTIVITY = 60 * 10

    def __init__(self):
        self.time_captured = None
        self._song_path = None
        self._song = None
        self._popen = None
        super().__init__()

        # making Baseplayer state -> private to public
        self.state = self._state

    def _pre_popen(self, song):
        return [
            "mpv",
            "--input-terminal=yes",
            "-no-audio-display",
            "--input-file=/proc/self/fd/0",
            "--no-term-osd",
            "--osc=no",
            "%s" % song,
        ]

    @threader_wrapper
    def _handle_pause_event(self):
        """
        Captures the time duration when the track is paused.
        Once track is unpause time will be added back to the original track time
        This time will be used to calculate the accuracy of current_position
        when pause state changes from pause to playing.
        """
        rollback = 0
        while True:
            rollback += 1
            if not self._track_paused or self._track_playing:
                self._track_start_time += rollback
                return
            sleep(1)  # must use sleep to prevent high cpu usage while threading

            if rollback > self._MAX_INACTIVITY:
                Verbose('Pydatpiff program was killed due to "Pause Inactivity"...')
                exit(1)

    @property
    def duration(self):
        """Return track length  in seconds"""
        if hasattr(self, "_metadata"):
            return self._metadata.track_duration
        return 0

    def _format_time(self, pos=None):
        """Format current song time to clock format"""
        pos = self.duration if not pos else pos
        minutes = int(pos / 60)
        seconds = int(pos % 60)
        return minutes, seconds

    @property
    def current_time(self):
        """Current time of track"""
        time_now = time() - self._track_start_time
        # always capture time to adjust pause time
        self.time_captured = time_now

        if self._track_stopped:
            return 0

        return time_now

    @current_time.setter
    def current_time(self, timer):
        if not self._track_stopped:
            self._track_start_time += timer

    def _sync_track_time(self, sec):
        """
        Adjust the track's time in seconds when track is
        being rewinding, fast-forwarding, or paused.
        """
        constrains = self._constrain_seek(sec)
        self.time_captured += constrains

    def _constrain_seek(self, seek):
        """Force range constraints on setting seek time,
        when rewinding and fast-forwarding.
        Time that is set out of range will be set to its nearest position.
        """
        seek = float(seek)
        current_pos = self.current_time

        # rewind past track's starting point
        if current_pos + seek < 0:
            return 0
        # ffwd past track's duration, then set track 5 seconds before ending.
        elif len(self) < current_pos + seek:
            # self.track_stopped = True
            return len(self) - 5.0

        return int(seek) * -1

    def _seek_syncer(func):
        """
        Callback function that force track time to be synced with the time
        whenever track position is being seeked see: MPV._seeker.
        """

        @wraps(func)
        def inner(self, time_sec):
            self._sync_track_time(time_sec)
            return func(self, time_sec)

        return inner

    @_seek_syncer
    def _seeker(self, sec=5):
        """
        Control fast-forward and rewind function.

        :param: pos - time to rewind or fast-forward (in seconds)
        """

        raw_sec = re.sub(r"-", "", str(sec))
        if not raw_sec.isnumeric():
            Verbose("Must use numerical numbers")
            return
        seek = "seek %s \n" % sec
        self._write_cmd(seek)
        return int(sec)

    def set_track(self, name, path):
        # media class method

        if File.is_file(path):
            self._song = name
            self._song_path = path
            self._metadata = MetaData(path)
            self._track_loaded = True
            self._track_start_time = time()
            self._volume = self._global_volume
            self.auto_manage_state()
        else:
            raise MvpError(1)

        Popen.unregister()

    def _write_cmd(self, cmd):
        """
        Handle user input and write command to media player( VLC or MPV) stdin.
        param: cmd - string character to write
        """
        if getattr(self, "_popen"):
            if self._track_loaded and self._popen.is_alive:
                self._popen.stdin.write("{}\n".format(cmd).encode("utf8"))
                self._popen.stdin.flush()
                return

    @property
    def play(self):
        # set_track method will handle the loadeding of track
        if self._track_loaded:
            if self._track_playing:
                return
            # if track not loaded, then load it and play
            self._popen = Popen(self._pre_popen(self._song_path))
            self._popen.register()
            self._track_loaded = True
            self._track_playing = True
            self.play  # noqa  - this is a property function in baseplayer
            self._volume = self._global_volume

    @property
    def pause(self):
        """Pause and unpause the track."""

        if self._track_loaded:
            if self._track_playing:
                cmd = "set pause yes \n"
                self._write_cmd(cmd)
                self._track_playing = False
                self._track_paused = True
                self._handle_pause_event()
            else:
                Verbose("\nUnpause")
                self._track_paused = False
                self._track_playing = True
                cmd = "set pause no \n"
                self._write_cmd(cmd)
        else:
            Verbose("No track playing")
        return

    def rewind(self, sec=5):
        """
        Rewind current track.
        :param: pos - time to rewind or fast-forward (in seconds)
        """
        sec = "-" + str(sec)
        self._seeker(sec)

    def ffwd(self, sec=5):
        """
        Fast-forward current track.
            :param: pos - time to rewind or fast-forward (in seconds)
        """
        sec = str(sec)
        self._seeker(sec)

    @property
    def stop(self):
        """Stop the current track from playing."""
        self._write_cmd("quit \n")
        self._track_stopped = True
        Popen.unregister()
        return

    @property
    def _volume(self):
        """Current media player volume"""
        return self._global_volume

    @_volume.setter
    def _volume(self, level):
        self._global_volume = level

    def volume(self, level=None):
        """Set the volume to exact number"""
        MAX_LEVEL = 200
        if not level or not isinstance(level, int):
            return

        if level > MAX_LEVEL:
            level = MAX_LEVEL

        elif level < 0:
            level = 0

        self._volume = level
        self._write_cmd("set volume %s" % level)

    def volume_up(self, vol=5):
        """Turn the media volume up"""

        if not isinstance(vol, int):
            return

        self._volume += vol

    def volume_down(self, vol=5):
        """Turn the media volume down"""

        if not isinstance(vol, int):
            return

        self._volume -= vol



class Error(Exception):
    __error__ = {
        1: "invalid mixtapes object",
        2: "no mixtapes found",
    }

    code = None
    message = "Unknown"

    def __init__(self, code, detail=""):
        self._code = code
        code = self.create(code, detail)
        super().__init__(code)

    def show(self, code):
        return self.__error__.get(code)

    @staticmethod
    def makeErrorName(error_code):
        # split and capitalize each name in error code
        error_name = "".join(list(map(lambda x: x.title(), error_code.split(" "))))

        if not (error_name.lower()).endswith("error"):
            error_name += "Error"
        return error_name

    @classmethod
    def create(cls, error_code, extra_message=""):
        if isinstance(error_code, str):
            max_errors = max(cls.__error__)
            cls.__error__[max_errors + 1] = Error.makeErrorName(error_code)
            return Error.create(max_errors + 1)

        elif error_code in cls.__error__:
            for code, error in cls.__error__.items():
                if not isinstance(error, str):
                    return error

                error_class_name = Error.makeErrorName(error)
                error_class = type(
                    error_class_name,
                    (cls,),
                    {"code": code, "message": error},
                )
                # apply the new error to the base error class
                # basically we are caching the new error
                cls.__error__[code] = error_class

            error_name = cls.__error__[error_code].__name__
            return "\n".join((error_name, extra_message))


class MixtapeError(Error):
    """handle all the Mixtape errors"""

    __error__ = {
        1: "Invalid category selected",
        2: "invalid input",
        3: "too few characters",
    }


class MediaError(Error):
    """handle all the Media errors"""

    __error__ = {
        1: "no mixtapes found",
        2: "invalid mixtapes object",
        3: "media album not set",
        4: "invalid track index",
        5: "song selection error",
        6: "unsupported media player",
        7: "media player not found",
        8: "song not found",
        9: "song unavailable",
    }


class AlbumError(Error):
    __error__ = {
        1: "Mixtape Not Found",
        2: "Invalid category selected",
    }


class Mp3Error(Error):
    __error__ = {
        1: "invalid player response",
        2: "album not found",
        3: "album name not found",
        4: "mp3 not found",
    }


class DatpiffError(Error):
    __error__ = {
        1: "Datpiff server down",
        2: "Datpiff desktop version failed",
        3: "Datpiff mobile version failed",
    }


class MvpError(Error):
    __error__ = {
        1: "song path does not exist",
    }


class PlayerError(Error):
    __error__ = {
        1: "Unsupported media object",
        2: "no song found",
        3: "derive class missing function",
        4: "call back method missing",
        5: "unsupported player",
        6: "player not found",
    }


class RequestError(Error):
    # _type = self.__class__._qualname__
    __error__ = {
        1: "datpiff server",
        2: "connection time out",
        3: "invalid url",
        4: "request error",
    }


class BuildError(Error):
    __error__ = {
        1: "user selection",
    }


class InstallationError(Error):
    _extra = (
        "\nPydatpiff Audio requires either VLC or MPV installation."
        "\n\nView"
        " https://github.com/cbedroid/pydatpiff/blob/master/README.md"
        " for more installation instructions."
        ""
    )
    __error__ = {
        1: "Pydatpiff installion error",
    }



import logging
import re
from functools import wraps

import bs4

from pydatpiff.constants import ampersands
from pydatpiff.errors import Mp3Error
from pydatpiff.urls import Urls
from pydatpiff.utils.request import Session

logger = logging.getLogger(__name__)


def escape_html_characters(char_list):
    results = []
    if not isinstance(char_list, (list, tuple)):
        char_list = [char_list]

    for char in char_list:
        if char in ampersands:
            amps = ampersands.index(char)
            char = re.sub(amps, "", char)
        results.append(char)
    return results


class MixtapeScraper:
    _MAX_RETRY = 5
    _MAX_MIXTAPES_PER_PAGE = 52  # maximum amount of mixtapes available per Datpiff's Page

    def __init__(self, base_response, limit=600):
        self._base_response = base_response  # Session.response
        self._soup = bs4.BeautifulSoup(base_response.text, "html.parser")

        # prepare request session
        self._session = Session()

        self._total_mixtapes = 0  # total mixtapes found
        # total mixtapes requested by user
        self._MIXTAPE_LIMIT = limit if isinstance(limit, int) else 520  # 10 pages

        self._initialize_attributes()
        self._get_page_links()

    @property
    def _attribute_list(self):
        """Mixtape dunder Attributes"""
        return [
            "_artists",
            "_mixtapes",
            "_links",
            "_ratings",
            "_views",
            "_album_covers",
        ]

    def _initialize_attributes(self):
        """Invoke  mixtape's attributes. See: pydatpiff.mixtapes.Mixtape"""
        for attr in self._attribute_list:
            setattr(self, attr, [])

    def _setMedias_attributes(self, bs4_content_list):

        for content in bs4_content_list.findAll(class_="contentItemInner"):
            # set album covers
            self._album_covers.extend([elem.find("img").get("src") for elem in content.findAll(class_="contentThumb")])

            # set artists
            self._artists.extend([elem.text for elem in content.findAll(class_="artist")])

            # Set mixtapes and links
            links = [elem.find("a") for elem in content.findAll(class_="title")]

            self._mixtapes.extend([re.sub(r"listen\sto", "", link.get("title")).strip() for link in links])
            self._links.extend([link.get("href") for link in links])

            # Set ratings and views
            self._ratings.extend([int(re.match(r"\d*", content.find(class_="text").img.get("alt"))[0])])
            self._views.extend(
                [
                    int(
                        re.sub(
                            r"\D*",
                            "",
                            content.findAll(class_="text")[1].span.text,
                        )[0]
                    )
                ]
            )

    @property
    def total_mixtapes(self):
        if hasattr(self, "_total_mixtapes"):
            return self._total_mixtapes
        return 0

    @total_mixtapes.setter
    def total_mixtapes(self, count):
        if isinstance(count, int):
            self._total_mixtapes += count

    def _parse_mixtape_page(self, url):
        """
        Mixtape's initiator function used to parse mixtape's page
        and returns detail used to set Mixtape's attributes.

        Args:
            url (str): mixtapes' page link url

        Returns:
            [int]: total number of mixtape's on page
        """
        text = self._session.method("GET", url=url).text
        # fmt: off
        content_container_id = "leftColumnWide"  # Datpiff Mixtape's main content wrapper
        content_wrapper_class = 'contentListing'
        mixtapes_content_items = "contentItem"  # Mixtape' Content Wrapper
        # fmt: on
        try:
            # Using BeautifulSoap
            soup = bs4.BeautifulSoup(text, "html.parser")
            content_container = soup.find(id=content_container_id)
            if content_container:
                content_listing = content_container.find(class_=content_wrapper_class)
                mixtape_items = content_listing.findAll(class_=mixtapes_content_items)
                # Set total mixtapes found
                self.total_mixtapes = len(mixtape_items)

                # set attribute for Mixtape Class
                self._setMedias_attributes(content_listing)
        except:
            logger.exception("CacheContentError")
            return 0

    def _get_page_links(self):
        """
        Return a list of html page links from mixtapes. Mixtape._select_mixtape method.
        """
        BASE_URL = Urls.datpiff["base"]

        # Check if pagination links are available
        pagination = self._soup.find(class_="pagination")
        if not pagination:
            # cache the first page and return the initial response url
            self._parse_mixtape_page(self._base_response.url)
            return [self._base_response.url]

        # Next get all pagination links anchor href.
        """
            Since this class (MixtapeScraper) has to be initialized with a mixtape, we should already have the
            content from the first page link (Active Page). Although we already processed this content,
            we still include it to accurately count to total mixtapes found. We should not be worried about recalling
            this request, since our `Session` will cache the response if it has already been requested.
        """
        page_links = pagination.find(class_="links").findAll("a")
        page_links = [BASE_URL + link.get("href") for link in page_links]
        for page_number, link in enumerate(page_links, start=1):
            # get the page link and parse it
            self._parse_mixtape_page(link)

            # If the max mixtapes is reached, then return the content
            # from the current page
            if self.total_mixtapes >= self._MIXTAPE_LIMIT:
                return page_links[:page_number]

        try:
            # If the max mixtapes is not reached, then return the last page of mixtapes
            return page_links  # return [<initial-url>]]
        except IndexError:
            # if all fails, then return the initial url
            self._parse_mixtape_page(self._base_response.url)
            return [self._base_response.url]

    def _request_get(self, url):
        """
        Thread safe request session's method.

        This method is used to get content from Datpiff's Mixtape web page
        during threading operation - ThreadQueue method.  See:  pydatpiff.backend.utils.ThreadQueue

        Args:
        url (str) - mixtape's link

        Returns:
            (str) - HTTP response text
        """
        return self._session.method("GET", url).text


class MediaScraper:
    @staticmethod
    def wrapper(f):
        @wraps(f)
        def inner(obj):
            try:
                f(obj)
            except:
                raise AttributeError("Regex Error")
            else:
                return f(obj)

        return inner

    @staticmethod
    def get_uploader_name(string):
        """Return the name of the person whom upload the mixtape"""
        try:
            return re.search(r'<a.*href="/profile.*>(.*)</a>', string).group(1)
        except AttributeError:
            pass
        return ""

    @staticmethod
    def get_uploader_bio(string):
        try:
            desc = re.findall(r'og:description".*content\="(.*)"', string)
            if desc:
                return escape_html_characters(desc[-1])[0].strip()
        except AttributeError:
            pass
        return ""

    @staticmethod
    def get_album_suffix_number(string):
        return re.search(r"\.(\d*)\.html", string).group(1)

    @staticmethod
    def get_embed_player_id(text):
        return re.search(r"/mixtapes/([\w\/]*)", text).group(1)  # noqa

    @classmethod
    def get_song_titles(cls, text):
        songs = re.findall(r'"title":"(.*\w*)",\s?"artist"', text)
        songs = list(escape_html_characters(songs))
        return songs

    @classmethod
    def get_duration_from(cls, text):
        return re.findall(r'"duration">(.*\d*)<', text)

    def get_mp3_urls(text):
        try:
            return re.findall(r"fix.concat\(\s\'(.*\w*)\'", text)  # noqa
        except AttributeError:
            raise Mp3Error(4)



# base player state management keys
# see: pydatpiff.backend.audio.baseplayer.BasePlayer.state
player_state_keys = {
    "loaded": "loaded",
    "playing": "playing",
    "paused": "paused",
    "stopped": "stopped",
    "system_stopped": "system_stopped",
}

# base player state symbols
music_symbols = {
    "playing": chr(9199),  # |>
    "paused": chr(9208),  # ||
    "stopped": chr(9209),  # []
    "music": chr(9836),  # music note
    "autoplay": chr(9871),  # loop
}


# Html unicode characters
ampersands = [
    "&quot;",  # ""
    "&amp;",  # &
    "&lt;",  # <
    "&gt;",  # >
    "&nbsp;",  # space
    "&iexcl;",  # ¡
    "&cent;",  # ¢
    "&pound;",  # £
    "&curren;",  # ¤
    "&yen;",  # ¥
    "&brvbar;",  # ¦
    "&sect;",  # §
    "&uml;",  # ¨
    "&copy;",  # ©
    "&ordf;",  # ª
    "&laquo;",  # «
    "&not;",  # ¬
    "&shy;",  #
    "&reg;",  # ®
    "&macr;",  # ¯
    "&deg;",  # °
    "&plusmn;",  # ±
    "&sup2",  #
    "&sup3;",  # ³
    "&acute;",  # ´
    "&micro;",  # µ
    "&para;",  # ¶
    "&middot;",  # ·
    "&cedil;",  # ¸
    "&sup1;",  # ¹
    "&ordm;",  # º
    "&raquo;",  # »
    "&frac14;",  # ¼
    "&frac12;",  # ½
    "&frac34;",  # ¾
    "&iquest;",  # ¿
    "&times;",  # ×
    "&divide;",  # ÷
    "&ETH;",  # Ð
    "&eth;",  # ð
    "&THORN;",  # Þ
    "&thorn;",  # þ
    "&AElig;",  # Æ
    "&aelig;",  # æ
    "&OElig;",  # Ø
    "&oelig;",  # ø
    "&Aring;",  # Å
    "&Oslash;",  # Ø
    "&Ccedil;",  # Ç
    "&ccedil;",  # ç
    "&szlig;",  # ß
    "&Ntilde;",  # Ñ
    "&ntilde;",  # ñ
]

verbose_message = {
    "MEDIA_INITIALIZED": "Media initialized",
    "MEDIA_SET": "Setting media to %s",
    "MEDIA_NOT_SET": "Media not set... Please call `setMedia` first",
    "SEARCH_SONG": "Searching for song: %s ...",
    "NO_SONG_SELECTED": "--- No song was entered --",
    "SONG_NOT_FOUND": "No song was found",
    "SONG_NAME_NOT_FOUND": "No song was found with the name: %s",
    "UNAVAILABLE_SONG": "This song is unavailable",
    "INVALID_DIRECTORY": "Invalid directory: %s",
    "SAVE_SONG": "Saving song: %s ...",
    "SAVE_ALBUM": "Album saved: %s to %s",
    "AUTO_PLAY_NO_SONG": "Play a song first to enable autoplay",
    "AUTO_PLAY_NEXT_SONG": "Playing next song",
    "AUTO_PLAY_LAST_SONG": "No more songs left to autoplay",
    "AUTO_PLAY_ENABLED": "\t----- AUTO PLAY ON -----",
    "AUTO_PLAY_DISABLED": "\t----- AUTO PLAY OFF -----",
    "AUTO_PLAY_INACTIVITY": "\t--- Autoplay stopped due to inactivity ---",
}

SERVER_DOWN_MSG = (
    "\n\t--- UNOFFICIAL DATPIFF MESSAGE --"
    "\nSorry, we encounter a problem with our server!"
    " Please try again later."
)



"""
    This file will stored all dynamic class and methods.
    These methods will be used throughout the whole program.

"""
import sys  # noqa: F401
import threading
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
from typing import List, Tuple, Union


def threader_wrapper(f):
    @wraps(f)
    def inner(*a, **kw):
        t = threading.Thread(target=f, args=(a), kwargs=dict(kw))
        t.daemon = True
        t.start()
        return t

    return inner


class ThreadQueue:  # pragma: no cover
    def __init__(self, main_job, input_work: Union[Tuple, List], *args, **kwargs):
        """
        This class will be used to execute concurrent jobs.
        The main job will be executed with the input work.
        The input work will be a list of work to be executed.
        :param input_work: input work to perform the main job with.
        """
        self.main_job = main_job  # job to perform with work
        self.input_work = input_work

    def execute(self, *args, **kwargs):
        """
        This method will execute the main job with the input work.
        args and kwargs are additional arguments kwargs
        :param args: args to pass to the main job.
        :param kwargs: kwargs to pass to the main job.

        """
        with ThreadPoolExecutor(max_workers=3) as executor:
            if args or kwargs:
                data = executor.map(
                    lambda work: self.main_job(work, *args, **kwargs), [work for work in self.input_work], timeout=10
                )
            else:
                data = executor.map(self.main_job, self.input_work)
            executor.shutdown(wait=True)
        return [x for x in data]


class Object:
    @staticmethod
    def is_dict(_type):
        return isinstance(_type, dict)

    @staticmethod
    def is_string(_type):
        return isinstance(_type, str)

    @staticmethod
    def is_list(_type):
        return isinstance(_type, (list, tuple))

    @staticmethod
    def strip_and_lower(string):
        """Return strip and lower value"""
        return str(string).lower().strip()

    @classmethod
    def remove_list_null_value(cls, _list):
        return list(filter(None, _list))

    @classmethod
    def enumerate_options(cls, data, start=0):
        """Return enumerate object"""
        if cls.is_dict(data) or cls.is_list(data):
            if cls.is_dict(data):
                data = data.items()
            return list(enumerate(data, start=start))
        raise NotImplementedError("datatype is not a dictionary or list ")


class Select:
    @classmethod
    def by_choices(cls, choice, options, fallback=None):
        """Select user choices and return the corresponding options
        Args:
            choice (str) - Expected choice
            options (list,tuple,dict):  Collection of options to select from.
            fallback (str,int, optional): Fallback choice if original expected choice
                                          is not found.
        """
        choice = Object.strip_and_lower(choice)
        val = None
        try:
            if Object.is_dict(options):
                val = min([key for key, val in options.items() if choice in Object.strip_and_lower(key)])

            elif Object.is_list(options):
                val = min([val for val in options if choice in Object.strip_and_lower(val)])

            elif Object.is_string(options):
                val = options if choice in Object.strip_and_lower(options) else fallback

            assert val is not None
        except ValueError:
            if fallback:  # no value found, then return fallback
                return cls.by_choices(choice=fallback, options=options, fallback=None)
            raise ValueError("No value found for {}".format(choice))
        return val

    @classmethod
    def get_leftmost_index(cls, index, options):
        """
        Select options by its index and return the left most index, or index minus 1.
        e.g. if index is 2 and options are [a,b,c,d,e] then return 1 (index 1).
               [a, b, c, d, e] , so index 2 = `c`, so leftmost is 1.

        """
        options_size = len(options)
        index -= 1
        index = 0 if (0 >= index or index > options_size) else index
        return index

    @classmethod
    def get_index_of(cls, choice, options):
        """
        Performs a per option, case-insensitive search on each option and return index of options that matches partial
        or whole string.
        e,g:
            options = ['apple', 'orange', 'grape']
            choice = 'org' or "orange"
            return 1

        :params:: choice,options
            choice - user choice. datatype: str
            options - list or dict object

        Returns:
            index of options
        """
        value = cls.by_choices(choice, options)
        if value:
            if Object.is_dict(options):
                options = list(options)
            return options.index(value)

        raise ValueError("Choice not found")



class Urls:
    datpiff = {
        "base": "https://www.datpiff.com/",
        "album": "http://www.datpiff.com",
        "search": "https://www.datpiff.com/mixtapes-search",
    }

    category = {
        "hot": "http://www.datpiff.com/mixtapes/hot",
        "new": "http://www.datpiff.com/mixtapes",
        "top": "http://www.datpiff.com/mixtapes-top",
        "celebrated": "http://www.datpiff.com/mixtapes/celebrated",
        "popular": "http://www.datpiff.com/mixtapes-popular.php",
        "exclusive": "http://www.datpiff.com/mixtapes-exclusive",
        "most download": "http://www.datpiff.com/mixtapes-popular.php?sort=downloads",
        "most listen": "http://www.datpiff.com/mixtapes-popular.php?filter=month&sort=listens",
        "most favorite": "http://www.datpiff.com/mixtapes-popular.php?sort=favorites",
        "highest rating": "http://www.datpiff.com/mixtapes-popular.php?filter=month&sort=rating",
    }

    @staticmethod
    def payload(artist):
        data = {
            "submit": "MTAxNTUuNzcxNTI5NDEyMzY0MTgwNzEx",
            "criteria": artist,
        }
        return data



from pydatpiff.utils.utils import Select

from .backend.scraper import MixtapeScraper
from .errors import MixtapeError
from .frontend.screen import Verbose
from .urls import Urls
from .utils.request import Session


class Mixtape(MixtapeScraper):
    valid_categories = list(Urls.category)
    _default_category = "hot"
    _user_selected = _default_category  # user  category or search input

    def __init__(self, category=None, search=None, limit=None, *args, **kwargs):
        """
        Mixtape Initialization.

        :param: category - name of the category to search from.
                            see Mixtape.category

        :param: search - search for an artist or mixtape's name
        """
        self._session = Session()
        self._session.clear_cache()

        self._select_mixtape(category=category, search=search)

        initial_page_content = self._request_response
        super().__init__(initial_page_content, limit=limit)

        if not len(self):
            Verbose("No Mixtape Found")
        else:
            Verbose("Found %s mixtapes" % len(self))

    def __str__(self):
        prefix = getattr(self, "_user_selected", self._default_category)
        return f"{prefix.title()} {self.__class__.__name__}"

    def __len__(self):
        if hasattr(self, "_artists"):
            if self._artists is not None:
                return len(self._artists)
        return 0

    @staticmethod
    def _validate_search(user_input):
        """Clean and force constraints on mixtape's search method.
        Args:
            user_input (str): user input
        """

        min_characters = 3
        if not isinstance(user_input, str):
            raise MixtapeError(2, "Expected datatype: string")

        if len(user_input) < min_characters:
            raise MixtapeError(
                3,
                """Not enough character: {}.\
                 Minimum characters limit is {}.""".format(
                    user_input, str(min_characters)
                ),
            )

        return user_input.strip()

    def _perform_search(self, name):
        """
        Search for an artist or mixtape's name.

        :param: name - name of an artist or mixtapes name
        """
        name = str(name).strip()
        Verbose("\nSearching for %s mixtapes ..." % name.title())
        url = Urls.datpiff["search"]
        return self._session.method("POST", url, data=Urls.payload(name))

    def _select_mixtape(self, category=None, search=None):
        """
        Initial setup. Gets all available mixtapes.

        :param: category - name of the category to search from.
                            (self Mixtape.category)
        :param: search - search for an artist or mixtape's name
        """
        if search:  # Search for an artist or mixtape
            filtered_search = self._validate_search(search)
            body = self._perform_search(filtered_search)
            self._user_selected = search  # capture user search input

        else:  # Selecting from category
            category = category or self._default_category
            if category.lower() not in self.valid_categories:
                category = self._default_category

            self._user_selected = category  # capture user category input
            choice = Select.by_choices(category, Urls.category)
            url = Urls.category[choice]  # get the url for the category
            body = self._session.method("GET", url)
        self._request_response = body
        return body

    @property
    def artists(self):
        """return all Mixtape artists' name"""
        if hasattr(self, "_artists"):
            return self._artists

    @property
    def album_covers(self):
        if hasattr(self, "_album_covers"):
            return self._album_covers

    @property
    def mixtapes(self):
        """Return all mixtapes name"""
        if hasattr(self, "_mixtapes"):
            return self._mixtapes

    @property
    def links(self):
        """Return all of the Mixtape' url links"""
        if hasattr(self, "_links"):
            return self._links

    @property
    def ratings(self):
        if hasattr(self, "_ratings"):
            return self._ratings

    @property
    def views(self):
        """Return the views count of each mixtapes"""
        if hasattr(self, "_views"):
            return self._views
